header {

package com.evelopers.unimod.analysis;

import java.util.*;
import com.evelopers.unimod.parser.*;
import com.evelopers.unimod.core.stateworks.*;
}

/**
 * Transforms AST to DNF tree
 */
class DNFTree extends TreeParser;
options {
    buildAST = true;    // uses CommonAST by default
    importVocab = Expr;
	defaultErrorHandler = false;
}

{
    public static AST optimize(AST ast) throws RecognitionException, InterpreterException {
        DNFTree tree = new DNFTree();
        tree.s(ast);
        tree.x(tree.getAST());
        tree.u(tree.getAST());
        tree.o(tree.getAST());
        return tree.getAST();
    }

    static AST toLiterals(AST ast) throws RecognitionException {
        DNFTree tree = new DNFTree();
        tree.s(ast);
        return tree.getAST();
    }

    static AST associativelyOpenBrackets(AST ast) throws RecognitionException {
        DNFTree tree = new DNFTree();
        tree.x(ast);
        return tree.getAST();
    }

    static AST toTerms(AST ast) throws RecognitionException {
        DNFTree tree = new DNFTree();
        tree.u(ast);
        return tree.getAST();
    }
}

s   : ! #(NOT child:s_neg) {#s = #child;}
    |   #(OR (s)*)
    |   #(AND (s)*)
    |   v
    ;

s_neg
    /* De Morgan transformations */
    : ! (AND | OR)
        {
            #s_neg = (#s_neg_in.getType() == AND) ? #[OR, "||"] : #[AND, "&&"];
            AST notNode = #[NOT,"!"];
            for (AST child = #s_neg_in.getFirstChild(); child != null; child = child.getNextSibling()) {
                notNode.setFirstChild(child);
                s(notNode);
                #s_neg.addChild(returnAST);
            }
        }

    /* Double negation */
    | ! #(NOT double_negated:s)         {#s_neg = #double_negated;}

    /* Negate relations */
    | ! #(EQUAL l_equal:t r_equal:t)    {#s_neg = #(#[NEQUAL, "!="], l_equal,  r_equal);}
    | ! #(NEQUAL l_nequal:t r_nequal:t) {#s_neg = #(#[EQUAL, "=="],  l_nequal, r_nequal);}
    | ! #(GE l_ge:t r_ge:t)             {#s_neg = #(#[LT, "<"],      l_ge,     r_ge);}
    | ! #(LE l_le:t r_le:t)             {#s_neg = #(#[GT, ">"],      l_le,     r_le);}
    | ! #(GT l_gt:t r_gt:t)             {#s_neg = #(#[LE, "<="],     l_gt,     r_gt);}
    | ! #(LT l_lt:t r_lt:t)             {#s_neg = #(#[GE, ">="],     l_lt,     r_lt);}

    /* Negate boolean constants */
    | ! TRUE  {#s_neg = #[FALSE,"false"];}
    | ! FALSE {#s_neg = #[TRUE, "true"];}

    /* Negate boolean variable */
    | ! var:t {#s_neg = #(#[NOT, "!"], #var);}
    ;

v   :   #(EQUAL t t)
    |   #(NEQUAL t t)
    |   #(GE t t)
    |   #(GT t t)
    |   #(LE t t)
    |   #(LT t t)

    |   (TRUE | FALSE)
    | ! IDENT {#v = astFactory.dup(#v_in);}
    ;

t ! :   (IDENT | CONST_NUM) {#t = astFactory.dup(#t_in);}
    ;

lit :   #(NOT v)
    |   v
    ;

x   : ! (AND | OR)
        {
            #x = #[x_in];
            for (AST child = #x_in.getFirstChild(); child != null; child = child.getNextSibling()) {
                x(child);
                /* Now returnAST has no k node in {&&, ||} that has k as its child */
                /* If returnAST has the same type as current node then add its children to current children list */
                #x.addChild(returnAST.getType() == #x_in.getType() ? returnAST.getFirstChild() : returnAST);
            }
        }
    |   lit
    ;

u   : ! AND
        {
            DistributiveBrackets distr = new DistributiveBrackets(astFactory);
            for (AST child = #u_in.getFirstChild(); child != null; child = child.getNextSibling()) {
                u(child);
                distr.and(returnAST);
            }
            #u = distr.getAST();
        }
    | ! OR
        {
            #u = #[OR];
            for (AST child = #u_in.getFirstChild(); child != null; child = child.getNextSibling()) {
                u(child);
                /* Join returned disjunction to current disjunction else add returned literal */
                #u.addChild(returnAST.getType() == OR ? returnAST.getFirstChild() : returnAST);
            }
        }
    |   lit
    ;

o   throws InterpreterException {Object skip;}
    : ! OR
        {   /* Optimize formula */
            ASTFormulaSimplifier formula = new ASTFormulaSimplifier(astFactory);
            for (AST child = #o_in.getFirstChild(); child != null; child = child.getNextSibling()) {
                formula.addTerm(o_term(child));
            }
            #o = formula.getSimplifiedFormula();
        }
    |   skip = o_term
    ;

o_term returns [ASTTermSimplifier term] throws InterpreterException
    { term = new ASTTermSimplifier(#o_term_in, astFactory); }
    : ! AND {#o_term = term.getSimplifiedTerm();}
    |   lit
    ;