header {
    package com.evelopers.unimod.analysis;

    import com.evelopers.unimod.analysis.executors.SubstitutionBuilder;
	import com.evelopers.unimod.parser.ConstNode;
}

/**
 * Gathers predicates and boolean variables. Set objectValues to 
 * constant nodes. Use case:
 * <pre><code>
 *   Gatherer gatherer = new Gatherer(typeInfoProvider);
 *   gatherer.gather(ast);
 *   SubstitutionBuilder builder = gatherer.getBuilder();
 *   Substitution substitution = builder.buildSubstitution();
 * </code></pre>
 */
class Gatherer extends TreeParser;
options {
    buildAST = false;
    importVocab = Expr;
	defaultErrorHandler = false;
}

{
	private TypeInfoProvider typeInfoProvider;		
	private SubstitutionBuilder builder;

	public Gatherer(TypeInfoProvider typeInfoProvider) {
		this();
		this.typeInfoProvider = typeInfoProvider;
		this.builder = new SubstitutionBuilder(typeInfoProvider);
	}		
	
	public SubstitutionBuilder getBuilder() {
		return builder;
	}
}

gather
    :   #(OR (gather)+)
    |   #(AND (gather)+)
    |   #(NOT gather)
    |   id:IDENT
        {
    		builder.addBooleanVar(id.getText());
    	}
    |   p:(EQUAL | NEQUAL | GE | GT | LE | LT) op1:(IDENT | CONST_NUM) op2:(IDENT | CONST_NUM)
        {
        	ConstNode constNode;
        	String variableName;
    		if (op1.getType() == IDENT) {
    			variableName = op1.getText();
    			constNode = (ConstNode) op2;
    		} else {
    			variableName = op2.getText();
    			constNode = (ConstNode) op1;
    		}
			builder.addPredicate(p.getType(), variableName, constNode.getText());
    		constNode.setObjectValue(typeInfoProvider.parseConstant(variableName, constNode.getText()));
        }
    |   (TRUE | FALSE)
    ;
